#!/usr/bin/env python3

"""hosts_resolver package entry point"""

import argparse
import json
import logging
import os
import sys
import typing
from socket import gaierror

from hosts_resolver.hosts_resolver import get_hosts_resolver


def main() -> typing.NoReturn:
    """Main entry point"""
    parser = argparse.ArgumentParser(
        description="Zero-dependency Python module to resolve names using /etc/hosts alone"
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s 1.1.0",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="enable debug mode",
    )
    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="output results as JSON",
    )
    parser.add_argument(
        "names",
        nargs="+",
        help="names to resolve using /etc/hosts",
    )

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    results: typing.Dict[str, typing.Optional[str]] = {}
    with get_hosts_resolver() as hosts_resolver:
        for name in args.names:
            try:
                results[name] = str(hosts_resolver.resolve(name))
            except gaierror:
                # if JSON output is disabled, exit fast to prevent inconsistent output
                if not args.json:
                    raise

                results[name] = None

    if args.json:
        print(json.dumps(results, indent=4 if args.debug else None))
    else:
        print(os.linesep.join(results.values()))  # type: ignore[arg-type]

    # exit with error if (at least) a resolution failed
    sys.exit(not all(results.values()))


if __name__ == "__main__":
    main()
