"""hosts_resolver module"""

import contextlib
import ctypes
import ipaddress
import logging
import typing
from socket import gaierror


class _Hostent(ctypes.Structure):  # pylint: disable=too-few-public-methods
    """hostent netdb.h structure ctypes mapping"""

    _fields_ = [
        ("h_name", ctypes.c_char_p),
        ("h_aliases", ctypes.POINTER(ctypes.c_char_p)),
        ("h_addrtype", ctypes.c_int),
        ("h_length", ctypes.c_int),
        ("h_addr_list", ctypes.POINTER(ctypes.POINTER(ctypes.c_ubyte))),
    ]


_IPAddress = typing.Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

_libc = ctypes.CDLL("libc.so.6")

# From <netdb.h> :
# struct hostent *gethostent(void);
_gethostent = _libc.gethostent
_gethostent.restype = ctypes.POINTER(_Hostent)
_gethostent.argtypes = ()

# void endhostent(void);
_endhostent = _libc.endhostent
_endhostent.restype = None
_endhostent.argtypes = ()


class _HostsResolver:
    """
    _HostsResolver context manager definition.

    Use this CM to resolve names using /etc/hosts entries, without any DNS request or subprocess
    call.

    Entries are fetched in a "lazy" way and cached to speed up processing against very large files.

    Known limitations:
      - `gethostent` is not reentrant (see GETHOSTBYNAME(3))
      - `gethostent` is not thread-safe (see GETHOSTBYNAME(3))
      - glibc implementations does not support IPv6 addresses (see GETHOSTBYNAME(3))
      - behavior is **undefined** when using multiple context managers simultaneously
    """

    def __init__(self) -> None:
        self._logger = logging.getLogger(self.__class__.__name__)
        self._cached_aliases: typing.Dict[str, _IPAddress] = {}

    def __enter__(self) -> "_HostsResolver":
        return self

    def __exit__(self, _, __, ___) -> None:
        _endhostent()

    def resolve(self, name: typing.Union[str, _IPAddress]) -> _IPAddress:
        """
        Try to resolve `name` using /etc/hosts entries.

        This method raises `socket.gaierror` when `name` is unknown.
        """
        # don't try to resolve IP addresses for idempotency !
        with contextlib.suppress(ValueError):
            return ipaddress.ip_address(name)

        while True:
            # fetch result from cache (if present)
            with contextlib.suppress(KeyError):
                return self._cached_aliases[name]  # type: ignore[index]

            # fetch "next" hosts entry
            entry_ptr = ctypes.cast(_gethostent(), ctypes.POINTER(_Hostent))
            if not entry_ptr:
                self._logger.debug("no more hosts entry")
                break

            # dereference structure pointer
            entry = entry_ptr.contents

            # decode entry "name"
            aliases = [entry.h_name.decode("ascii")]

            self._logger.debug("got new hosts entry: %s", aliases[0])

            # decode entry "aliases"
            i = 0
            h_aliases = ctypes.cast(entry.h_aliases, ctypes.POINTER(ctypes.c_char_p))
            while h_aliases[i]:
                aliases.append(h_aliases[i].decode("ascii"))
                self._logger.debug("found new alias for %s: %s", aliases[0], aliases[-1])
                i += 1

            # decode "network addresses" from raw bytes array
            addresses = ctypes.cast(
                entry.h_addr_list,
                ctypes.POINTER(ctypes.POINTER(ctypes.c_ubyte * entry.h_length)),
            )
            address = ipaddress.ip_address(bytes(addresses[0].contents))

            # populate cache with all entry aliases
            for alias in aliases:
                self._logger.debug("caching %s as %s...", address, alias)
                self._cached_aliases[alias] = address

        raise gaierror(f"could not resolve: {name}")


@contextlib.contextmanager
def get_hosts_resolver() -> typing.Generator[_HostsResolver, None, None]:
    """Return an `_HostsResolver` instance to use as a context manager"""
    with _HostsResolver() as hosts_resolver:
        yield hosts_resolver
